namespace StudySystemClient.Response {
	public class Response {
		public int transaction_id { get; private set; }
		public Body body { get; private set; }

		public Response.from_bytes(uint8[] bytes) throws DecodeError {
			Der.Sequence sequence;
			try {
				sequence = Der.decode(bytes) as Der.Sequence;
			} catch (Der.DecodeError e) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response was not valid DER: " + e.message);
			}
			if (sequence == null) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response was not a SEQUENCE");
			}
			if (sequence.value.length != 2) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response sequnce contained %u fields (expected 2)",
					sequence.value.length);
			}
			var id_datum = sequence.value[0] as Der.Integer;
			if (id_datum == null) {
				throw new DecodeError.INVALID_RESPONSE(
					"Response transactionId was not an INTEGER");
			}
			transaction_id = (int)id_datum.value;
			body = Body.from_datum(sequence.value[1]);
		}
	}

	public errordomain DecodeError {
		INVALID_BODY,
		INVALID_RESPONSE,
		NOT_IMPLEMENTED,
	}

	public abstract class Body {
		protected enum Tag {
			ERROR = 0,
			ACK = 1,
			PRIORITIZED_ACTIVITIES = 2,
		}

		internal static Body from_datum(Der.Datum datum) throws DecodeError {
			var choice = datum as Der.Choice;
			if (choice == null) {
				throw new DecodeError.INVALID_BODY(
					"ResponseBody was not a CHOICE");
			}
			switch (choice.id) {
			case Tag.ERROR:
				return new Error.from_datum(choice.value);
			case Tag.ACK:
				return new Ack.from_datum(choice.value);
			case Tag.PRIORITIZED_ACTIVITIES:
				throw new DecodeError.NOT_IMPLEMENTED(
					"PrioritizedActivities not yet implemented");
			default:
				throw new DecodeError.INVALID_BODY(
					"Invalid ResponseBody tag");
			}
		}
	}

	public class Error : Body {
		public enum Value {
			INVALID_REQUEST = 0,
			INVALID_ARGUMENTS = 1,
			SERVER_ERROR = 2,
		}

		public Value value { get; private set; }

		internal Error.from_datum(Der.Datum datum) throws DecodeError {
			var enumerated = datum as Der.Enumerated;
			if (enumerated == null) {
				throw new DecodeError.INVALID_BODY(
					"Error was not an ENUMERATED");
			}
			if (enumerated.value < 0 || enumerated.value > 2) {
				throw new DecodeError.INVALID_BODY(
					"Error type was not in range 0..2");
			}
			value = (Value)enumerated.value;
		}
	}

	public class Ack : Body {
		internal Ack.from_datum(Der.Datum datum) throws DecodeError {
			var @null = datum as Der.Null;
			if (@null == null) {
				throw new DecodeError.INVALID_BODY("Ack was not NULL");
			}
		}
	}
}
